class Column:

    def __init__(self, name):
        self.name = name       #<- カラム名

    def __eq__(self, value):   #<- 「==」演算子を上書き
        return Expression(self, "==", value)  #<- 式を返す

    def __ne__(self, value):   #<- 「!=」演算子を上書き
        return Expression(self, "!=", value)  #<- 式を返す

class Expression:     #<- 式を表すクラス

    def __init__(self, left, op, right):
        self.left  = left        #<- 演算子の左側
        self.op    = op          #<- 演算子
        self.right = right       #<- 演算子の右側

class QueryBuilder:   #<- SQLを生成するクラス

    def __init__(self, table):
        self._table = table      #<- テーブル名
        self._wheres = []        #<- where句の条件式

    def where(self, expr):       #<- where句に相当
        self._wheres.append(expr)
        return self

    def to_sql(self):      #<- SQLを生成
        buf = []
        buf.extend(("select * from ", self._table))
        if self._wheres:         #<- where句があれば
            buf.append(" where ")   #<- 条件式をSQLに変換
            for expr in self._wheres:
                self._expr2sql(expr, buf)
                buf.append(" and ")
            buf.pop()            #<- 余分な「 and 」を外す
        return "".join(buf)

    def _expr2sql(self, expr, buf):
        buf.append(self._sql_literal(expr.left))
        sql_op = self._sql_op(expr.op, expr.right)
        buf.extend((" ", sql_op, " "))
        buf.append(self._sql_literal(expr.right))

    def _sql_op(self, op, right):
        if op == "==":           #<- 「==」演算子は、
            if right is None:    #<- 右側がNoneなら
                return "is"      #<- 「is」に変換し、
            else:                #<- それ以外なら
                return "="       #<- 「=」に変換する。
        if op == "!=":           #<- 「!=」演算子は、
            if right is None:    #<- 右側がNoneなら
                return "is not"  #<- 「is not」に変換し、
            else:                #<- それ以外なら
                return "<>"      #<- 「<>」に変換する。
        raise TypeError("%r: operator not supported." % (op,))

    def _sql_literal(self, val):
        if val is None:  return "null"
        if val is True:  return "true"
        if val is False: return "false"
        if isinstance(val, int): return str(val)
        if isinstance(val, str): return self._quote(val)
        if isinstance(val, Column): return val.name
        raise TypeError("%r: unknown data type." % (val,))

    def _quote(self, s):
        s = s.replace("'", "''")         #<- 「'」をエスケープしてから
        return f"'{s}'"                  #<- 「'」で囲う

## 使い方：Pythonの式をSQLに変換する
id = Column('id')
sql = QueryBuilder("users").where(id == 100).to_sql()
print(sql)    #=> select * from users where id = 100
## 使い方：where句に複数の条件式を指定する
sql = (QueryBuilder("users")
          .where(Column('id') == 100)
          .where(Column('deleted_at') == None)
          .to_sql())
print(sql)
  #=> select * from users where id = 100 and deleted_at is null
