import sys
from functools import wraps
from types import FunctionType

task_funcs = {}    #<- キー：タスク名、値：タスク関数

def task(name=None):
    def deco(func, name=name):  #<- 新しい高階関数を生成
        task_name = name or func.__name__
        task_funcs[task_name] = func   #<- タスク関数を登録
        return func
    if isinstance(name, FunctionType):  #<- 引数が関数なら
        func = name
        return deco(func, None)  #<- 関数を返す
    else:                   #<- 引数が関数でなければ
        return deco         #<- 生成した高階関数を返す

def main(args=None):
    if args is None:
        args = sys.argv[1:]  #<- argv[0]はファイル名なので除外
    task_names = args
    for task_name in task_names:
        try:
            task_func = task_funcs[task_name]
        except KeyError:    #<- タスク関数がなければエラー
            msg = f"task '{task_name}' not defined."
            sys.stderr.write(f"error: {msg}\n")
            sys.exit(1)
        task_func()         #<- タスク関数を実行
    sys.exit(0)

def run(command):           #<- UNIXコマンドを実行するヘルパー関数
    print(command)          #<- （のつもり、今は表示するだけ）
    #print(f"$ {command}")  #<- （またはプロンプトをつける）

## 関数を最初の1回だけ呼び出し、2回目以降はその結果を再利用する
def once(func):
    done   = False          #<- 1回でも実行するとTrueにする
    result = None           #<- func()の実行結果を保持
    @wraps(func)
    def newfunc():
        nonlocal done, result
        if not done:        #<- まだ実行してないなら
            result = func() #<-   func()を実行して結果を保持
            done = True     #<-   実行したのでTrueにする
        return result       #<- 何度呼び出しても同じ結果を返す
    return newfunc

## 使い方
@task
@once      #<- 共通する初期化作業は1回だけ行えばいい
def prepare():
    print("... do something ...")

@task
def compile():
    prepare()
    run("uplatex foo.tex")

@task
def makepdf():
    prepare()
    run("dvipdfmx foo.dvi")

@task("all")
def task_all():
    compile()      #-> compile()中のprepare()が呼ばれる
    makepdf()      #-> makepdf()中のprepare()が呼ばれる

main()
