from functools import wraps
from types import FunctionType
from unittest import TestCase

## テストメソッドにつける関数デコレータ（用の高階関数）
def fixture(func):
    ## すでに関数デコレータが使われていれば、もとの関数を探す
    fn = func
    while hasattr(fn, '__wraps__'):
        fn = fn.__wraps__
    ## 引数がselfしかなければ、何もしない
    argcount = fn.__code__.co_argcount
    argnames = fn.__code__.co_varnames[1:argcount]
    if not argnames:
        return func
    ## self以外の位置引数があれば、新しい関数を作成して返す
    @wraps(func)
    def deco(self):
        ## フィクスチャ用のメソッドまたは関数を探して実行し、
        ## その戻り値を収集する
        fixture_objs = []
        for name in argnames:
            ## メソッドがあればそれを、なければグローバル関数を探す
            fixture_func = getattr(self, f"fixture_{name}", None) or \
                           fn.__globals__.get(f"fixture_{name}", None)
            if fixture_func is None:
                raise NameError(f"{name}: fixture not found.")
            fixture_objs.append(fixture_func())
        ## self以外の引数にはフィクスチャデータを渡す
        return func(self, *fixture_objs)
    return deco

## テストクラスにつけるクラスデコレータ（用の高階関数）
def fixtures(klass):
    ## テストメソッドの一覧
    pairs = []
    for name, val in klass.__dict__.items():
        if name.startswith("test") and isinstance(val, FunctionType):
            pairs.append((name, val))
    ## テストメソッドに関数デコレータを適用する
    for name, func in pairs:
        setattr(klass, name, fixture(func))
    ## クラスオブジェクトをそのまま返す
    return klass

## 使用例（関数デコレータ）
class Dummy1Test(TestCase):

    def fixture_user(self):
        return {"id": 2000, "name": "Alice"}

    def fixture_admin_user(self):
        return {"id": 100, "name": "Bob"}

    @fixture
    def test_1(self, user):
        self.assertEqual(user["name"], "Alice")

    @fixture
    def test_2(self, admin_user):
        self.assertEqual(admin_user["name"], "Bob")

## 使用例（クラスデコレータ）
@fixtures
class Dummy2Test(TestCase):

    def fixture_user(self):
        return {"id": 2000, "name": "Alice"}

    def fixture_admin_user(self):
        return {"id": 100, "name": "Bob"}

    def test_1(self, user):
        self.assertEqual(user["name"], "Alice")

    def test_2(self, admin_user):
        self.assertEqual(admin_user["name"], "Bob")
