import re

class Template:
    """テンプレートクラス"""

    def __init__(self, string, filename=None, escapefunc=None):
        self.filename = filename
        self._escapefunc = escapefunc or self.ESCAPE_FUNC
        self.program  = self.parse(string)
        self._pycode  = None

    ESCAPE_FUNC  = ""
    BLOCK_REXP   = re.compile(r'^([ \t]*)%%% ')
    KEEPING_REXP = re.compile(r'^(?:elif\s|else:|except[\s:]|finally:)')
    ENDING_REXP  = re.compile(r'^#end')
    INLINE_REXP  = re.compile(r'\$\{(.*?)\}')

    def parse(self, string):
        """テンプレート文字列をPythonプログラムに変換する"""
        buf = []
        self._parse_block(string, buf)
        return "".join(buf)

    def _parse_block(self, string, buf):
        """「%%% 文」をパースする"""
        block_rexp = self.BLOCK_REXP
        keep_rexp  = self.KEEPING_REXP
        end_rexp   = self.ENDING_REXP
        indent = ""
        for line in string.splitlines(True):
            if block_rexp.match(line):
                line = block_rexp.sub('\\1', line)
                buf.append(line)
                ## ・「if:」や「for:」ならインデントを増やす
                ## ・「#end」ならインデントを減らす
                ## ・「elif」や「else:」ならインデントはそのまま
                line = line.lstrip()
                if end_rexp.match(line):
                    indent = indent[:-4]
                elif keep_rexp.match(line):
                    pass
                else:
                    indent += "    "
            else:
                s = self._parse_inline(line)
                buf.append(f"{indent}_buf.append(f'{s}')\n")

    def _parse_inline(self, line):
        """「${式}」をパースする"""
        pos = 0
        buf = []
        rexp = self.INLINE_REXP
        for m in rexp.finditer(line):
            expr = m.group(1)
            text = line[pos:m.start()]
            pos  = m.end()
            if text:
                buf.append(self._escape_text(text))
            if expr:
                buf.append(self._escape_expr(expr))
        remained = line[pos:]
        if remained:
            buf.append(self._escape_text(remained))
        return "".join(buf)

    def _escape_text(self, text):
        return text.replace("\\", r"\\").replace("'", r"\'")\
                   .replace("{", "{{").replace("}", "}}")\
                   .replace("\n", r"\n")

    def _escape_expr(self, expr):
        return "{%s(%s)}" % (self._escapefunc, expr)

    def render(self, vars):
        """テンプレートを実行し、その結果（文字列）を返す"""
        if self._pycode is None:
            filename = self.filename or "<string>"
            self._pycode = compile(self.program, filename, "exec")
        vars['_buf'] = []
        exec(self._pycode, vars, vars)
        _buf = vars.pop('_buf')
        return "".join(_buf)

## 使い方
string = """
<div>
  <ul>
%%% for i, x in enumerate(items, 1):
    %%% if i % 2:
    <li class="odd">${x}</li>
    %%% else:
    <li class="even">${x}</li>
    %%% #end
%%% #end
  </ul>
</div>
"""[1:]
filename = "example.html"
t = Template(string, filename, escapefunc=None)
print("-------- program --------")
print(t.program)
print("-------- output --------")
vars = {
    "items": [10, 20, 30],
}
output = t.render(vars)
print(output)
