import sys
import ast


OPTABLE = {
    ## クラス : 演算子
    ast.Eq    : "==",
    ast.NotEq : "!=",
    ast.Gt    : ">",
    ast.GtE   : ">=",
    ast.Lt    : "<",
    ast.LtE   : "<=",
    ast.Is    : "is",
    ast.IsNot : "is not",
    ast.In    : "in",
    ast.NotIn : "not in",
}

def op2str(op):
    """比較演算子の文字列を返す"""
    opstr = OPTABLE.get(type(op), None)
    assert opstr, "unknown op: %r" % (op,)
    return opstr


class AssertionTransformer(ast.NodeTransformer):

    def visit_Assert(self, node):
        """assert文に第2引数を追加する"""
        ## assert文の第2引数があれば何もせずreturn
        if node.msg is not None:
            return node
        ## assert文の第1引数が比較式であれば
        test_node = node.test
        if isinstance(test_node, ast.Compare):
            ## 必要な情報を取り出す
            op     = test_node.ops[0]
            format = "%r {} %r".format(op2str(op))
            right  = test_node.comparators[0]
            elems  = [test_node.left, right]
            ## 比較演算子が連続していた場合も対応
            i = 1
            while i < len(test_node.ops):
                op = test_node.ops[i]
                format += " {} %r".format(op2str(op))
                right = test_node.comparators[i]
                elems.append(right)
                i += 1
        ## assert文の第1引数が比較式でなければ
        else:
            elems = [test_node]
            format = "%r"
        ## assert文の第2引数を追加
        def cl(node, test_node=test_node):
            return ast.copy_location(node, test_node)
        tuple_node = ast.Tuple(elts=elems, ctx=ast.Load())
        msg_node = ast.BinOp(op=cl(ast.Mod()),
                             left=cl(ast.Str(s=format)),
                             right=cl(tuple_node))
        node.msg = cl(msg_node)
        ## assert文のノードを返す
        return node


class Main:

    def load_module(self, filename):
        ## *.py を読み込む
        with open(filename, encoding='utf-8') as f:
            source = f.read()
        ## パースして構文解析木を得て、改変してからコンパイル
        node = ast.parse(source)
        new_node = AssertionTransformer().visit(node)
        code = compile(new_node, filename, 'exec')
        ## モジュールオブジェクトを作成
        modname = filename.replace(".py", "").replace("/", ".")
        mod = type(sys)(modname)
        mod.__dict__['__file__']   = filename
        mod.__dict__['__module__'] = modname
        eval(code, mod.__dict__, mod.__dict__)
        return mod

    def run_modules(self, modules):
        from unittest import TestLoader, TestSuite, TextTestRunner
        loader = TestLoader()
        tests = []
        for mod in modules:
            tests.extend(loader.loadTestsFromModule(mod))
        suite = TestSuite(tests)
        return TextTestRunner(verbosity=2).run(suite)

    def run(self, *filenames):
        testmodules = [ self.load_module(fname) for fname in filenames ]
        result = self.run_modules(testmodules)
        return result

    @classmethod
    def main(cls, args=None):
        if args is None: args = sys.argv[1:]
        result = cls().run(*args)
        sys.exit(int(not result.wasSuccessful()))


if __name__ == '__main__':
    Main.main()
